#!/usr/bin/env python
# Author: Roberto Rodriguez (@Cyb3rWard0g)
# License: BSD 3-Clause
import json
import requests
import tarfile
from urllib.parse import urlparse
from pathlib import Path
import pandas as pd
from pyspark.sql import SparkSession
from openhunt.logparser import winlogbeat

def get_spark():
    return (SparkSession.builder.appName("Mordor").config("spark.sql.caseSensitive", "True").getOrCreate())

def decompressJSON(filePath):
    tf = tarfile.open(filePath)
    for member in tf.getmembers():
        tf.extract(member)
        tf.close()
        return member.name

def downloadFile(url, dest="/tmp/"):
    response = requests.get(url, stream=True)
    outFilePath = "{}{}".format(dest, (Path(url).resolve().name))
    if response.status_code == 200:
        with open(outFilePath, 'wb') as f:
            f.write(response.raw.read())
    return outFilePath

# Function to read a json file
def readJSON(filePath, spark=False):
    if (spark):
        df = spark.read.json(filePath)
    else:
        df= pd.read_json(filePath, lines = True)
    return df
    
# Function to parse winlogbeat data (all versions)
def processDataFrame(df, spark=False):
    if (spark):
        print("[+] Processing a Spark DataFrame..")
        if '@metadata' in df.columns:
            df = df.withColumn('version', df["@metadata.version"].substr(1,1))
            df = df.withColumn('beat_type', df["@metadata.beat"])
            win = winlogbeat()
            # Verify what verion of Winlogbeat was used to ship the data
            if (len(df.filter((df.beat_type == 'winlogbeat') & (df.version <= 6)).limit(1).take(1)) > 0 ):
                df = win.winlogbeat_6(df.filter((df.beat_type == 'winlogbeat') & (df.version <= 6)), 'Spark')
            elif (len(df.filter((df.beat_type == 'winlogbeat') & (df.version >= 7)).limit(1).take(1)) > 0 ):
                df = win.winlogbeat_7(df.filter((df.beat_type == 'winlogbeat') & (df.version >= 7)),'Spark')
    else:
        print("[+] Processing Pandas DataFrame..")
        if '@metadata' in df.columns:
            df['version'] = df['@metadata'].apply(lambda x : x.get('version'))
            df['version'] = df['version'].astype(str).str[0]
            df['beat_type'] = df['@metadata'].apply(lambda x : x.get('beat'))
            # Initialize Empty Dataframe
            df = pd.DataFrame()
            win = winlogbeat()
            # Verify what verion of Winlogbeat was used to ship the data
            if ((df['beat_type'] == 'winlogbeat') & (df['version'] <= '6')).any():
                version_6_df = win.winlogbeat_6(df[(df['beat_type'] == 'winlogbeat') & (df['version'] <= '6')], 'Pandas')
                df = df.append(version_6_df, sort = False)              
            if ((df['beat_type'] == 'winlogbeat') & (df['version'] >= '7')).any():
                version_7_df = win.winlogbeat_7(df[(df['beat_type'] == 'winlogbeat') & (df['version'] >= '7')], 'Pandas')
                df = df.append(version_7_df, sort = False)                      
            if (df['beat_type'] != 'winlogbeat').any():
                not_winlogbeat = df[df['beat_type'] != 'winlogbeat']
                df = df.append(not_winlogbeat, sort = False)            
            df = df.dropna(axis = 0,how = 'all').reset_index(drop = True)
    print("[+] DataFrame Returned !")
    return df


#!/usr/bin/env python

# Author: Jose Rodriguez (@Cyb3rPandaH)
# License: GNU General Public License v3 (GPLv3)

import pandas as pd
import requests
import yaml

def getDictionary(platform = '', provider = '', event = '', url = None):
    ''' 
    If platform = 'aws' --> do not change default value of 'provider'
    If provider is an ETW PROVIDER such as 'Microsoft-Windows-Security-Auditing' --> insert platform = 'windows/etw-providers'
    For right format of platform, proviser and event check: https://github.com/hunters-forge/OSSEM/tree/master/source/data_dictionaries
    This function returns a dictionary
    '''
    if provider == 'security':
        provider = 'Microsoft-Windows-Security-Auditing'

    if url == None:
        urlBase = 'https://raw.githubusercontent.com/hunters-forge/OSSEM/master/source/data_dictionaries/'
        url = urlBase + platform + '/' + provider + '/events/'+event+'.yml'
        ossemFile = requests.get(url)
        ossemDict = yaml.safe_load(ossemFile.text)
        return ossemDict
    else:
        ossemFile = requests.get(url)
        ossemDict = yaml.safe_load(ossemFile.text)
        return ossemDict


def getEventDf(platform = '', provider = '', event = '', url = None):
    ''' 
    If platform = 'aws' --> do not change default value of 'provider'
    If provider is an ETW PROVIDER such as 'Microsoft-Windows-Security-Auditing' --> insert platform = 'windows/etw-providers'
    For right format of platform, proviser and event check: https://github.com/hunters-forge/OSSEM/tree/master/source/data_dictionaries
    This function returns a pandas dataframe
    '''
    dictionary = getDictionary(platform, provider, event, url)
    eventFields = dictionary.get('event_fields')
    eventCode = dictionary.get('event_code')
    eventTitle = dictionary.get('title')
    dfBase = pd.DataFrame(columns=['event_code','title','standard_name','standard_type','name','type','description','sample_value'])

    for i in eventFields:
        valuesToAdd = list(i.values())
        valuesToAdd.insert(0,eventCode)
        valuesToAdd.insert(1,eventTitle)
        seriesToAdd = pd.Series(valuesToAdd, index = dfBase.columns)
        dfBase = dfBase.append(seriesToAdd,ignore_index = True)
    return dfBase

def getEventFields(platform = '', provider = '', event = '', url = None):
    '''
    We are using the unique() method because in some OSSEM dictionaries the name of a field is repeated.
    This function returns a List
    '''
    df = getEventDf(platform, provider, event, url)
    eventFields = df['name'].drop_duplicates().to_list()
    return eventFields


#!/usr/bin/env python

# Author: Jose Rodriguez (@Cyb3rPandaH)
# License: GNU General Public License v3 (GPLv3)

import pandas as pd

class winlogbeat(object):
        # Function to parse winlogbeat data up to version 6
        def winlogbeat_6(self, df, df_type):
                print("[+] Processing Data from Winlogbeat version 6..")
                if df_type == 'Pandas':
                        # Extract event_data nested fields
                        event_data_field = df['event_data'].apply(pd.Series)
                        df = df.drop('event_data', axis = 1)
                        df_flat = pd.concat([df, event_data_field], axis = 1)
                        df= df_flat\
                                .dropna(axis = 1,how = 'all')\
                                .rename(columns={'log_name':'channel','record_number':'record_id','source_name':'provider_name'})
                elif df_type == 'Spark':
                        df = df.select('event_data.*','source_name','log_name','record_number','event_id','computer_name','@timestamp','message')
                        df = df\
                                .withColumnRenamed("log_name", "channel")\
                        .withColumnRenamed("source_name", "provider_name")\
                        .withColumnRenamed("record_number","record_id")
                else:
                        exit
                return df
        # Function to parse winlogbeat data since version 7
        def winlogbeat_7(self, df, df_type):
                print("[+] Processing Data from Winlogbeat version 7..")
                if df_type == 'Pandas':
                        winlog_field = df['winlog'].apply(pd.Series)
                        event_data_field = winlog_field['event_data'].apply(pd.Series)
                        df_flat = pd.concat([df, winlog_field, event_data_field], axis = 1)
                        df= df_flat.dropna(axis = 1,how = 'all').drop(['winlog','event_data','process'], axis = 1)
                        df['level'] = df['log'].apply(lambda x : x.get('level'))
                elif df_type == 'Spark':
                        df = df.select('winlog.event_data.*','winlog.channel','winlog.provider_name','winlog.record_id','winlog.event_id','winlog.computer_name','@timestamp','message')
                        df = df\
                                .withColumnRenamed("winlog.channel", "channel")\
                        .withColumnRenamed("winlog.provider_name", "provider_name")\
                        .withColumnRenamed("winlog.record_id","record_id")\
                                .withColumnRenamed("winlog.event_id","event_id")\
                                .withColumnRenamed("winlog.computer_name","computer_name")
                else:
                        exit
                return df


#!/usr/bin/env python
# Author: Roberto Rodriguez (@Cyb3rWard0g)
# License: BSD 3-Clause

from openhunt.utils import *

import requests
from io import BytesIO
from zipfile import ZipFile

def downloadMordorFile(url, dest="/tmp/"):
    mordorFile = downloadFile(url)
    mordorJSONPath = decompressJSON(mordorFile)
    return mordorJSONPath
    
def registerMordorSQLTable(spark, url, tableName):
    mordorJSONPath = downloadMordorFile(url)
    mordorDF = readJSON(mordorJSONPath, spark)
    mordorDF = processDataFrame(mordorDF, spark)
    mordorDF.createOrReplaceTempView(tableName)
    print("[+] Temporary SparkSQL View: {} ".format(tableName))

def getMordorZipFile(url):
    '''
    The initial idea for this function is here: https://stackoverflow.com/questions/5710867/downloading-and-unzipping-a-zip-file-without-writing-to-disk
    '''
    url = url + '?raw=true'
    zipFileRequest = requests.get(url)
    zipFile = ZipFile(BytesIO(zipFileRequest.content))
    jsonFilePath = zipFile.extract(zipFile.namelist()[0])
    return jsonFilePath


#!/usr/bin/env python

# Author: Jose Rodriguez (@Cyb3rPandaH)
# License: GNU General Public License v3 (GPLv3)

import pandas as pd


import matplotlib.pyplot as plt, seaborn as sns

from pyspark.sql import DataFrame
from pyspark.sql.types import *
import pyspark.sql.functions as func

    
# Function to get a horizontal bar chart with data labels
def barh_chart(dataframe,xfield,yfield,title,xlabel = '',ylabel = '',figSize = (12,8)):
    '''
    I did not create this function. You can find its reference here: 
    https://stackoverflow.com/questions/28931224/adding-value-labels-on-a-matplotlib-bar-chart
    '''
    # Bring some raw data.
    if isinstance(dataframe,pd.DataFrame) == True:
        frequencies = dataframe[xfield].values[::-1].tolist()
        max_freq = dataframe[xfield].values.max()
        min_freq = dataframe[xfield].values.min()
        y_labels = dataframe[yfield].values[::-1].tolist()
    elif isinstance(dataframe, DataFrame) == True:
        frequencies = dataframe.toPandas()[xfield].values[::-1].tolist()
        max_freq = dataframe.toPandas()[xfield].values.max()
        min_freq = dataframe.toPandas()[xfield].values.min()
        y_labels = dataframe.toPandas()[yfield].values[::-1].tolist()

    freq_series = pd.Series(frequencies)

    # Plot the figure.
    plt.figure(figsize = figSize)
    ax = freq_series.plot(kind='barh')
    ax.set_title(title)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_yticklabels(y_labels)
    ax.set_xlim(min_freq, max_freq) # expand xlim to make labels easier to read
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(True)
    ax.spines['left'].set_visible(True)

    rects = ax.patches

    # For each bar: Place a label
    for rect in rects:
        # Get X and Y placement of label from rect.
        x_value = rect.get_width()
        y_value = rect.get_y() + rect.get_height() / 2

        # Number of points between bar and label. Change to your liking.
        space = 5
        # Vertical alignment for positive values
        ha = 'left'

        # If value of bar is negative: Place label left of bar
        if x_value < 0:
            # Invert space to place label to the left
            space *= -1
            # Horizontally align label at right
            ha = 'right'

        # Use X value as label and format number with one decimal place
        label = "{:,.0f}".format(x_value)

        # Create annotation
        plt.annotate(
            label,                      # Use `label` as label
            (x_value, y_value),         # Place label at end of the bar
            xytext=(space, 0),          # Horizontally shift label by `space`
            textcoords="offset points", # Interpret `xytext` as offset in points
            va='center',                # Vertically center label
            ha=ha)                      # Horizontally align label differently for
                                        # positive and negative values.  
    
# Function to get a histogram
def histogram(dataframe,field):
     # Parameters:
        # datadrame: must be a spark dataframe
        # field: String value. It must match a single column name
    # About nulls values: This function does not consider null values on its calculations
    # Selecting column. Dropping null values.
    df=dataframe.select(field).dropna(how='any',subset=field)
    # Converting spark dataframe into pandas dataframe so we can use Seaborn library to create our box plot
    panda_df=df.toPandas()
    return sns.distplot(panda_df[field], kde = False) # This function returns a box plot
    
    
# Function to get a box plot
def box_plot(dataframe,field):
     # Parameters:
        # datadrame: must be a spark dataframe
        # field: String value. It must match a single column name
    # About nulls values: This function does not consider null values on its calculations
    # Selecting column. Dropping null values.
    df=dataframe.select(field).dropna(how='any',subset=field)
    # Converting spark dataframe into pandas dataframe so we can use Seaborn library to create our box plot
    panda_df=df.toPandas()
    return sns.boxplot(x=panda_df[field]) # This function returns a box plot

# Function to get a time series line chart
def time_series_chart(dataframe,field,interval,operation,data=False,fig_width=16,fig_height=4):
     # Parameters:
        # datadrame: must be a pandas dataframe
        # field: String value. It must match a single column name. The function will group the results using the field column
        # interval: String value. Find aliases here: https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
        # operation: String value. Operations defined so far: 'count','sum','mean'
        # data: Boolean value. Default value is False. When True, the function also returns a pandas dataframe with data used in the chart
        # fig_width: Integer value. Default value is 16. This parameter affects the final width of the chart
        # fig_height: Integer value. Default value is 4. This parameter affects the final heigh of the chart
    # About nulls values: This function does not consider null values on its calculations
    # Importing Libraries and Modules
    import pyspark.sql.functions as F
    import matplotlib.pyplot as plt
    from matplotlib import dates as mpl_dates
    from datetime import datetime, timedelta
    # Applying seaborn style for chart
    plt.style.use('seaborn')
    # Using conditional statement based on 'operation' parameter
    if operation=='count':
        # Grouping index (timestamp) and field
        analysis=dataframe.groupby([field,pd.Grouper(freq=interval)])[field].count()
    elif operation=='sum':
        # Grouping index (timestamp) and field
        analysis=dataframe.groupby([field,pd.Grouper(freq=interval)])[field].sum()
    elif operation=='mean':
        # Grouping index (timestamp) and field
        analysis=dataframe.groupby([field,pd.Grouper(freq=interval)])[field].mean()
    else:
        print('Operation has not been defined yet')
        return # Execution ends here

    # Unstacking dataframe by field. All columns will have the same index start
    analysis_plot=analysis.unstack(field)
    # Plotting time series line chart
    analysis_plot.plot(linestyle='-',figsize=(fig_width,fig_height),title=operation+' of '+field+' every '+interval)
    # Auto formatting of dates axis (Inclination of axis labels)
    plt.gcf().autofmt_xdate()
    # Setting format of dates axis
    date_format = mpl_dates.DateFormatter('%D %T')
    # Applying format to dates axis
    plt.gca().xaxis.set_major_formatter(date_format)
    # Setting format of chart legend
    ax=plt.gca()
    plt.legend(bbox_to_anchor=(1.1, 1.1), bbox_transform=ax.transAxes)
    if data==True:
        return analysis_plot.transpose().fillna('').dropna(axis=1,how='all') # This function also shows a pandas dataframe with details by field and time interval

# Function to get a horizontal bar chart with data labels
def time_series_dataframe(dataframe,time_field):
     # Parameters:
        # datadrame: must be a spark dataframe
        # field: String value. It must match a single column name
    # About nulls values: This function does not consider null values on its calculations
    # Importing Libraries and Modules
    import pyspark.sql.functions as F
    # Dropping null values. Creating a new column with timestamp format. Dropping original time_field. Converting spark dataframe into pandas dataframe. Setting 'timestamp' as dataframe index. Sorting new index
    df = dataframe.dropna(how='any',subset = time_field).withColumn('timestamp',F.to_utc_timestamp(time_field,tz = 'UTC')).drop(time_field).toPandas().set_index('timestamp').sort_index()
    return df # This functions returns a pandas dataframe


#!/usr/bin/env python

# Author: Jose Rodriguez (@Cyb3rPandaH)
# License: GNU General Public License v3 (GPLv3)

import pandas as pd
from pyspark.sql import DataFrame
from pyspark.sql.types import *
import pyspark.sql.functions as func
from pyspark.sql.functions import col, avg, min, max, stddev, skewness, kurtosis
   
def summary(dataframe):
    '''
    Identifying dataframe type, count of fields, and count of records.
    This function only identifies pandas and spark dataframes 
    '''
    if isinstance(dataframe, pd.DataFrame) == True: # Valdiating if object is a pandas dataframe
        dataframeType = 'Pandas Dataframe'
        countFields = dataframe.shape[0] # Calculating number of columns
        countRecords = dataframe.shape[1] # Calculating number os rows
    elif isinstance(dataframe, DataFrame) == True:
        dataframeType = 'Spark Dataframe'
        countFields = len(dataframe.columns)
        countRecords = dataframe.count()
    else:
        dataframeType = 'Other'
        countFields = 'Not supported by this function'
        countRecords = 'Not supported by this function'
    
    # Printing summary of dataframe information
    print('Dataframe type             : ', dataframeType)
    print('Number of fields available : ', countFields)
    print('Number of records available: ', countRecords)

def numStats(dataframe, field):
    '''
    This function works with pandas and spark dataframe
    Parameter field must be a String value, and it must make reference to a single column name
    Parameter field must make reference to a numerical variable
    This function does not consider null values on its calculations
    '''
    if isinstance(dataframe, pd.DataFrame) == True:
        df = dataframe[field].dropna() # Selecting column and droping null values
        # Count of Values
        count = df.count()
        countNullValues = dataframe.shape[0] - count
        # Central Tendency
        mean_value = df.sum() / count
        median_value = df.median()
        # Min, max, and Percentiles
        min_value = df.min()
        max_value = df.max()
        percentile_25 = df.quantile(0.25)
        percentile_75 = df.quantile(0.75)
        # Variation
        stddev_value = df.std()
        range_value = max_value - min_value
        IQR_value = percentile_75 - percentile_25
        #Shape
        skewness_value = df.skew()
        kurtosis_value = df.kurt()
    elif isinstance(dataframe, DataFrame) == True:
        df = dataframe.select(field).dropna(how='any')
        # Count of Values
        count=df.count()
        countNullValues = dataframe.select(field).count() - count
        # Central Tendency
        mean_process=df.agg(avg(col(field))) # The result of agg is a spark dataframe
        mean_value=mean_process.collect()[0][mean_process.columns[0]] # The result of collect is a list of row. [0] first element & [mean_process.columns[0] name of the column
        median_value=df.approxQuantile(col=field,probabilities=[0.5],relativeError=0.05)[0] # The result of approxQuantile is a list. [0] first element
        # Min, Max, and Percentiles
        min_process=df.agg(min(col(field))) # The result of agg is a spark dataframe
        min_value=min_process.collect()[0][min_process.columns[0]] # The result of collect is a list of row. [0] first element & [min_process.columns[0] name of the column
        max_process=df.agg(max(col(field))) # The result of agg is a spark dataframe
        max_value=max_process.collect()[0][max_process.columns[0]] # The result of collect is a list of row. [0] first element & [max_process.columns[0] name of the column
        percentile_25=df.approxQuantile(col=field,probabilities=[0.25],relativeError=0.05)[0] # The result of approxQuantile is a list. [0] first element
        percentile_75=df.approxQuantile(col=field,probabilities=[0.75],relativeError=0.05)[0] # The result of approxQuantile is a list. [0] first element
        # Variation
        stddev_process=df.agg(stddev(col(field))) # The result of agg is a spark dataframe
        stddev_value=stddev_process.collect()[0][stddev_process.columns[0]] # The result of collect is a list of row. [0] first element & [stddev_process.columns[0] name of the column
        range_value=max_value-min_value # Calculation of the range of values
        IQR_value=percentile_75-percentile_25 # Calculation of the Interquartile range
        # Shape
        skewness_process=df.agg(skewness(col(field))) # The result of agg is a spark dataframe
        skewness_value=skewness_process.collect()[0][skewness_process.columns[0]] # The result of collect is a list of row. [0] first element & [skewness_process.columns[0] name of the column
        kurtosis_process=df.agg(kurtosis(col(field))) # The result of agg is a spark dataframe
        kurtosis_value=round(kurtosis_process.collect()[0][kurtosis_process.columns[0]],2) # The result of collect is a list of row. [0] first element & [kurtosis_process.columns[0] name of the column
    
    
    # Printing summary of statistics
    print('Summary of Descriptive Statistics - ',field)
    print('**********************************************************')
    print('Count of values          : ',count)
    print('Count of Null values     : ',countNullValues)
    print('Central tendency:-----------------------------------------')
    print('Mean(Average)            : ',round(mean_value,2))
    print('Median(Percentile 25)    : ',round(median_value,2))
    print('Min, Max, and Percentiles:--------------------------------')
    print('Minimum                  : ',round(min_value,2))
    print('Maximum                  : ',round(max_value,2))
    print('Percentile 25 (Q1)       : ',round(percentile_25,2))
    print('Percentile 75 (Q3)       : ',round(percentile_75,2))
    print('Variation:------------------------------------------------')
    print('Standard Deviation       : ',round(stddev_value,2))
    print('Range                    : ',round(range_value,2))
    print('Interquartile Range (IQR): ',round(IQR_value,2))
    print('Shape:----------------------------------------------------')
    print('Skewness                 : ',round(skewness_value,2))
    print('Kurtosis                 : ',round(kurtosis_value,2))
    print('**********************************************************')
    # Creating a dictionary with descriptive statistics
    data = {'Statistic': ['count', 'Count Null Values', 'mean', 'median', 'min', 'max', 'percentile25', 'percentile75', 'stddev', 'range', 'IQR', 'skewness', 'kurtosis'],
            'Values': [count,countNullValues,mean_value,median_value,min_value,max_value,percentile_25,percentile_75,stddev_value,range_value,IQR_value,skewness_value,kurtosis_value]}
    # Creating a pandas dataframe
    summary_stats = pd.DataFrame(data)
    return summary_stats # This function returns a pandas dataframe

# Function to obtain a spark dataframe with stack counting operation over a field
def stack_count(dataframe, field, ascending = False):
    # Parameters:
        # datadrame: spark dataframe
        # field: String or List value. It must match column names. When field is a list, the function counts different fields combinations
    # About nulls values: This function does not consider null values on its calculations
    # Importing Libraries and Modules
        # Nothing to import
    
    if isinstance(dataframe,pd.DataFrame) == True:
        print('Pandas dataframe: Still working on it!! Coming Soon!!')
    elif isinstance(dataframe, DataFrame) == True:
        # Selecting column. Dropping null values.
        df = dataframe.select(field).dropna(how='any',subset=field)
        # Grouping by field using count function. The result is a spark dataframe
        df_count = df.groupBy(field).count()
        # Sorting spark dataframe by count column
        df_sorted = df_count.orderBy('count',ascending = ascending)
        # Printing message that indicates the number of rows in the spark dataframe. By default, the show() method only shows a maximum of 20 rows
        # Adding a column with relative frequency
        total_count = df.count()
        df_final = df_sorted.withColumn('%', func.round(df_sorted['count'].cast(IntegerType())*100/total_count,1))
        print('IMPORTANT!! The result contains ',df_final.count(),' rows.')
        return df_final # This function returns a spark dataframe
    else:
        print('This function only supports Spark dataframes')

# Function to obtain the upper and lower limits for the +/- 1.5(IQR) rule. This values are used to determine if a value within a spark dataframe column can be considered an outlier
def iqr_outliers_limits(self, dataframe, field):
    # Paramenters:
        # datadrame: must be a spark dataframe
        # field: String value. It must match a single column name
    # About nulls values: This function does not consider null values on its calculations
    # Importing Libraries and Modules
        # Nothing to import
    # Selecting column. Dropping null values
    df=dataframe.select(field).dropna(how='any',subset=field)
    # Applying approxQuatile function to calculate percentiles 25,50, and 75. The result is a list with the three values
    quartiles_list=df.approxQuantile(col=field,probabilities=[0.25,0.5,0.75],relativeError=0.05)
    # Calculating the interquartile range, which is equal to Percentile75 - Percentile25. The interquartile range is a statistic to measure variability in a set of numerical data
    IQR=quartiles_list[2]-quartiles_list[0]
    # The lower level is equal to Percentile25 (first element of quartiles_list: index 0) - 1.5 times IQR
    lower_limit=quartiles_list[0]-1.5*IQR
    # The upper level is equal to Percentile75 (third element of quartiles_list: index 2) + 1.5 times IQR
    upper_limit=quartiles_list[2]+1.5*IQR
    # Creating a list with 2 elements: lower level and upper level
    limits_list = [lower_limit,upper_limit]
    return limits_list # This function returns a list

# Function to add a column with outlier/normal tag to a spark dataframe based on IQR rule (+/- 1.5 IQR)
def outlier_tag(self,dataframe,field,filter=False):
    # Paramenters:
        # datadrame: must be a spark dataframe
        # field: String value. It must match a single column name
        # filter: Boolean value. Default value is False. When False, the function returns the original spark dataframe with new column 'outlier_tag'. When True, the function returns the original spark dataframe with new column 'outlier_tag' and filtered by 'outlier_tag' using the value 'outlier'
    # About nulls values: This function does not consider null values on its calculations
    # Importing Libraries and Modules
    import pyspark.sql.functions as F
    # Dropping rows where field column shows null values
    df=dataframe.dropna(how='any',subset=field)
    # Calling iqr_outliers_limits function
    limits_IQR_rule = self.iqr_outliers_limits(df,field)
    # Adding new column to the dataframe with tag 'outlier' or 'normal'
    df_tagged=df.withColumn('outlier_tag',F.when((F.col(field)<limits_IQR_rule[0]) | (F.col(field)>limits_IQR_rule[1]), 'outlier').otherwise('normal'))
    # Filtering dataframe by 'outlier_tag' column using value 'outlier'
    df_filtered=df_tagged.filter(df_tagged.outlier_tag == 'outlier')
    # Using conditional statement based on 'filter' parameter
    if filter==True:
        new_df=df_filtered
    else:
        new_df=df_tagged
    # Counting number of rows in df_filtered ('outlier')
    outliers_qty=df_filtered.count()
    # Printing message that indicates the number of outliers in the dataframe. By default, the show() method only shows a maximum of 20 rows
    print('There are ',outliers_qty,' outliers within ',field,' values')
    return new_df # This function returns a spark dataframe






#!/usr/bin/env python

# Author: Jose Rodriguez (@Cyb3rPandaH)
# License: GNU General Public License v3 (GPLv3)

import setuptools

with open('README.md')as f:
    long_description = f.read()

setuptools.setup (
	name = 'openhunt',
	version = '1.6.8',
	author = 'Jose Rodriguez',
	description = 'A Python library to expedite the analysis of data during hunting engagements',
	long_description=long_description,
	long_description_content_type="text/markdown",
	url="https://github.com/Cyb3rPanda/openhunt",
	keywords="threat hunting python pandas jupyter notebook",
	packages=setuptools.find_packages(),
	install_requires=[
        'pandas',
		'altair',
		'seaborn',
		'matplotlib',
		'pyspark'
    	],
	license='GNU General Public License v3 (GPLv3)',
	classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Topic :: Security',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
		'Programming Language :: Python :: 3.7',
    ]
)
